use std::{env::temp_dir, net::SocketAddr};

use tokio::io::AsyncWriteExt;
use tokio_stream::StreamExt;
use tonic::transport::Server;
use uuid::Uuid;

use crate::{
    app::SharedApp,
    services::release_manager::{
        extensions::ReleaseManagerExt, models::CommitArtifact, ReleaseManager,
    },
};

use self::gen::{
    flux_releaser_server, CommitArtifactRequest, CommitArtifactResponse, UploadArtifactRequest,
    UploadArtifactResponse,
};

pub mod gen {
    tonic::include_proto!("flux_releaser");
}

pub struct FluxReleaserGrpc {
    release_manager: ReleaseManager,
}

impl FluxReleaserGrpc {
    pub fn new(app: SharedApp) -> Self {
        Self {
            release_manager: app.release_manager(),
        }
    }
}

#[tonic::async_trait]
impl flux_releaser_server::FluxReleaser for FluxReleaserGrpc {
    async fn upload_artifact(
        &self,
        request: tonic::Request<tonic::Streaming<UploadArtifactRequest>>,
    ) -> std::result::Result<tonic::Response<UploadArtifactResponse>, tonic::Status> {
        let mut stream = request.into_inner();

        let file_path = temp_dir()
            .join("flux_releaser")
            .join("tmp")
            .join("upload_artifact")
            .join(Uuid::new_v4().to_string());
        tokio::fs::create_dir_all(file_path.parent().unwrap()).await?;
        let mut file = tokio::fs::File::create(&file_path).await?;

        while let Some(item) = stream.next().await {
            tracing::trace!("received chunk");
            let item = item?;

            let _ = file.write(&item.content).await?;
        }

        file.flush().await?;

        let upload_id = match self.release_manager.upload_artifact(file_path.into()).await {
            Ok(res) => res,
            Err(e) => return Err(tonic::Status::unknown(e.to_string())),
        };

        Ok(tonic::Response::new(UploadArtifactResponse {
            upload_id: upload_id.to_string(),
        }))
    }

    async fn commit_artifact(
        &self,
        request: tonic::Request<CommitArtifactRequest>,
    ) -> std::result::Result<tonic::Response<CommitArtifactResponse>, tonic::Status> {
        let req = request.into_inner();
        let artifact = self
            .release_manager
            .commit_artifact(
                req.try_into()
                    .map_err(|e: anyhow::Error| tonic::Status::invalid_argument(e.to_string()))?,
            )
            .await
            .unwrap();

        Ok(tonic::Response::new(CommitArtifactResponse {
            artifact_id: artifact.to_string(),
        }))
    }
}

impl TryFrom<CommitArtifactRequest> for CommitArtifact {
    type Error = anyhow::Error;

    fn try_from(value: CommitArtifactRequest) -> Result<Self, Self::Error> {
        if value.app.is_empty() {
            anyhow::bail!("app cannot be empty")
        }
        if value.branch.is_empty() {
            anyhow::bail!("branch cannot be empty")
        }
        if value.upload_id.is_empty() {
            anyhow::bail!("folder cannot be empty")
        }

        Ok(Self {
            app: value.app,
            branch: value.branch,
            upload_id: value.upload_id.try_into()?,
        })
    }
}

pub async fn tonic_serve(host: SocketAddr, app: SharedApp) -> anyhow::Result<()> {
    tracing::info!("grpc listening on: {}", host);
    Server::builder()
        .add_service(flux_releaser_server::FluxReleaserServer::new(
            FluxReleaserGrpc::new(app),
        ))
        .serve(host)
        .await?;

    Ok(())
}
