use std::path::PathBuf;

use crate::{app::infra::grpc::FluxReleaserGrpcClient, grpc::gen::CommitArtifactRequest};

use super::{
    archive::Archive,
    file_reader::FileReader,
    flux_releaser_uploader::FluxReleaserUploader,
    release_manager::models::{ArtifactID, UploadArtifactID},
};

pub struct FluxLocalClusterManager {
    file_reader: FileReader,
    archive: Archive,
    flux_uploader: FluxReleaserUploader,
    flux_releaser_client: FluxReleaserGrpcClient,
}

impl FluxLocalClusterManager {
    pub fn new(
        file_reader: FileReader,
        archive: Archive,
        flux_uploader: FluxReleaserUploader,
        flux_releaser_client: FluxReleaserGrpcClient,
    ) -> Self {
        Self {
            file_reader,
            archive,
            flux_uploader,
            flux_releaser_client,
        }
    }

    pub async fn package_clusters(
        &self,
        include: impl Into<PathBuf>,
    ) -> anyhow::Result<UploadArtifactID> {
        let files = self.file_reader.read_files(include.into()).await?;
        let archive = self.archive.create_archive(files).await?;
        let upload_id = self.flux_uploader.upload_archive(archive).await?;

        Ok(upload_id)
    }

    pub async fn commit_artifact(
        &self,
        app: impl Into<String>,
        branch: impl Into<String>,
        upload_id: UploadArtifactID,
    ) -> anyhow::Result<ArtifactID> {
        let artifact_id = self
            .flux_releaser_client
            .lock()
            .await
            .commit_artifact(tonic::Request::new(CommitArtifactRequest {
                app: app.into(),
                branch: branch.into(),
                upload_id: upload_id.to_string(),
            }))
            .await?;

        artifact_id.into_inner().artifact_id.try_into()
    }
}

pub mod extensions {
    use crate::{
        app::SharedLocalApp,
        services::{
            archive::extensions::ArchiveExt, file_reader::extensions::FileReaderExt,
            flux_releaser_uploader::extensions::FluxReleaserUploaderExt,
        },
    };

    use super::FluxLocalClusterManager;

    pub trait FluxLocalClusterManagerExt {
        fn flux_local_cluster_manager(&self) -> FluxLocalClusterManager;
    }

    impl FluxLocalClusterManagerExt for SharedLocalApp {
        fn flux_local_cluster_manager(&self) -> FluxLocalClusterManager {
            FluxLocalClusterManager::new(
                self.file_reader(),
                self.archive(),
                self.flux_releaser_uploader(),
                self.grpc_client.clone(),
            )
        }
    }
}
